/*
 * Copyright 2003 Amber Archer Consulting Co., Inc. All rights reserved.
 * AMBER ARCHER PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package org.amberarcher.xml.tool.highlight;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringReader;
import java.io.Writer;
import java.net.MalformedURLException;
import java.net.URL;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.w3c.dom.Element;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;

/**
 * A highlighter that uses the <a href="http://www.gnu.org/software/src-highlite/">GNU Src-Highlight</a>
 * package to render Java, C/C++, Prolog, Perl, PHP3, Python, Flex, ChangeLog or Ruby source code as
 * XHTML.
 *
 * @author <a href="mailto:kdowney@amberarcher.com">Kyle F. Downey</a>
 * @version $Revision: 1.4 $
 */
public class SourceHighlighter implements Highlighter {
    public Node highlight(Node replaceNode) throws HighlightException {
        Reader srcReader = getSourceCodeReader(replaceNode);
        String lang = replaceNode.getAttributes().getNamedItem("lang").getNodeValue();
        try {
            Runtime rt = Runtime.getRuntime();
            Process p = rt.exec(new String[]{"source-highlight", "--out-format=xhtml", "--src-lang=" + lang, "--tab=4",
                    "--css=highlight.css"});

            HighlightFeeder feeder = new HighlightFeeder(srcReader, p);
            HighlightOutputReader outReader = new HighlightOutputReader(p.getInputStream());

            feeder.start();
            outReader.start();

            int exitCode = p.waitFor();
            if (exitCode > 0) {
                throw new HighlightException("source-highlight process failed: exit code " + exitCode);
            }

            Document highlightedDoc = parseHighlightedSrc(outReader.getSource());
            return highlightedDoc.getDocumentElement().getElementsByTagName("pre").item(0);
        } catch (IOException e) {
            throw new HighlightException("Failed to create source-highlight process", e);
        } catch (InterruptedException e) {
            throw new HighlightException("Error waiting on source-highlight process", e);
        }
    }

    public void setHighlightContext(HighlightContext ctx) {
        this.ctx = ctx;
    }

    private Reader getSourceCodeReader(Node replaceNode) throws HighlightException {
        Node hrefNode = replaceNode.getAttributes().getNamedItem("href");
        if (null == hrefNode) {
            StringBuffer txt = new StringBuffer();
            NodeList children = replaceNode.getChildNodes();
            for (int ii = 0; ii < children.getLength(); ii++) {
                Node child = children.item(ii);
                if (!((child instanceof Text) || (child instanceof CDATASection))) {
                    throw new HighlightException("cannot highlight non-text content");
                } else {
                    txt.append(child.getNodeValue());
                }
            }


            return new StringReader(txt.toString());
        } else {
            try {
                String href = hrefNode.getNodeValue();
                URL srcUri = null;

                try {
                    srcUri = new URL(href);
                } catch (MalformedURLException e) {
                    srcUri = ctx.resolveRelativeUri(href);
                }

                return new InputStreamReader(srcUri.openStream());
            } catch (MalformedURLException e) {
                throw new HighlightException("bad source href: " + hrefNode.getNodeValue(), e);
            } catch (IOException e) {
                throw new HighlightException("failed to retrieve source: " + hrefNode.getNodeValue(), e);
            }
        }
    }

    private static Document parseHighlightedSrc(String src) throws HighlightException {
        try {
            DocumentBuilder docBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
            return docBuilder.parse(new InputSource(new StringReader(src)));
        } catch (Exception e) {
            throw new HighlightException("failed to parse XHTML output tree", e);
        }
    }

    private static final void copyStreams(Reader reader, Writer writer) throws IOException {
        int count;
        char[] chunk = new char[4096];

        while ((count = reader.read(chunk)) >= 0) {
            writer.write(chunk, 0, count);
        }
    }

    private static class HighlightFeeder extends Thread {
        public HighlightFeeder(Reader srcReader, Process p) {
            this.srcReader = srcReader;
            this.process = p;
        }

        public void run() {
            try {
                Writer procWriter = new BufferedWriter(new OutputStreamWriter(process.getOutputStream()));
                copyStreams(new BufferedReader(srcReader), procWriter);
                procWriter.flush();
                procWriter.close();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        private Reader srcReader;
        private Process process;
    }

    private static class HighlightOutputReader extends Thread {
        public HighlightOutputReader(InputStream is) {
            this.is = is;
        }

        public void run() {
            try {
                InputStreamReader isr = new InputStreamReader(is);
                BufferedReader br = new BufferedReader(isr);
                String line = null;
                while ((line = br.readLine()) != null) {
                    src.append(line);
                    src.append("\n");
                }
            } catch (IOException ioe) {
                ioe.printStackTrace();
            }
        }

        public String getSource() {
            return src.toString();
        }

        private InputStream is;
        private StringBuffer src = new StringBuffer();

    }

    private HighlightContext ctx;
}
